#ifndef CTS_DISTRIBUTION__
#define  CTS_DISTRIBUTION__
#include "random.h"
#include "util.h"
using namespace::std;
//ctsdistribution is the base class for univatvariate continuous distributions
//it is a pure base class so you can never have variables
// with this class, only pointers to them or references.
class ctsdistribution {
public:
  //  constructors and destructors
  ctsdistribution(double a, double b=0., rng &tg=globalrandom) :g(tg) {par=new double[2];par[0]=a;par[1]=b;}
    ctsdistribution(const ctsdistribution &a):g(a.g){par=new double[2];par[0]=a.par[0];par[1]=a.par[1];}
    virtual ~ctsdistribution(){ delete[] par;};
// something like a virtual constructor
// Visual c++ has trouble with an identical situation to Stroustrup p424....
    virtual ctsdistribution *clone()=0;
//  functions for sampling
	double operator()() {return next();};//get a random number like a function
	dvec sample(int n){return next(n);}
	double sample(){return next();};
	dvec next(int n){dvec x(*this,n);return x;};
//  densities - and cumulative densities
	virtual double log_pdf(double x) =0;
	double diff_log_pdf(double x, double y) {return log_pdf(x)-log_pdf(y);};
//  functions for output and input
	virtual void print(void) const =0;	
	virtual ostream &print(ostream &o) const=0;
	friend istream &operator>>(istream &in, ctsdistribution *x) {x=scan(in);return in;}
	friend ctsdistribution *scan(istream &in);
	friend ctsdistribution *parscan(istream &in, const char *namestring,char default_val[]="undefined");
        virtual ctsdistribution & operator=(ctsdistribution &a) {
		if (&a==this) return a;
		par[0]=a.par[0];par[1]=a.par[1];g=g;
		return *this;
	}
	virtual bool undefined(){ return false;}
protected:	
	double *par;
	int np;
	rng &g;
	virtual double next()=0;
private: 
	ctsdistribution( rng &tg=globalrandom) :g(tg){};//try to make sure this is never called
};

class gammad: public ctsdistribution {
public :
        ctsdistribution* clone(){return new gammad(*this);}
    	ostream &print(ostream &o) const {
		o << "gamma" << "(" << par[0]<< "," << par[1] << ")";
		return o;
	}
	void print(void) const {
		cout << "gamma"<< "(" << par[0]<< "," << par[1] << ")";
	}
	double shape(void) const {return par[0];}
	double scale(void) const {return par[1];}
	double log_pdf(double x) { 
		double lx = x*scale();
		return (shape()-1.)*log(lx) -lgamma(shape()) - lx + log(scale());
	}
	gammad(double shape, double scale=1., rng &tg=globalrandom) 
		: ctsdistribution(shape,scale,tg) {}
	gammad(const gammad &c):ctsdistribution(c){};
//	~gammad(){};
private:
	double next() {return g.rgamma(shape(),1./scale());}
};
//
//  a constant class for completeness
//
class constant: public ctsdistribution {
public :

	 ctsdistribution* clone(){return new constant(*this);}
ostream &print(ostream &o) const {
		o << "constant(" << value() <<")"; 
		return o;
	}
	void print(void) const {
		cout << "constant(" << value() <<")"; 
	}
	double value(void) const {return par[0];} 
	double log_pdf(double x) { 
		return 0.0;
	}
	constant(double val):ctsdistribution(val){};
	constant(const constant &val):ctsdistribution(val){}; 
	~constant(){};
	double next(){return value();}
};
//
//  normal random numbers
//
class normal: public ctsdistribution {
public :
ctsdistribution* clone(){return new normal(*this);}
	ostream &print(ostream &o) const {
		o << "normal("<<mu()<<","<<sigma()<<")";
		return o;
	}
	void print(void) const {
		cout << "normal("<<mu()<<","<<sigma()<<")";
	}
	double mu(void) const {return par[0];}
	double sigma(void) const {return par[1];}
	double log_pdf(double x) {
		double lx = (x-mu())/sigma();
		return LMY_SQRT_2PI - log(sigma()) - 0.5*lx*lx;
	}
	normal(double mu=0.0, double sigma=1,rng &tg=globalrandom)
		:ctsdistribution(mu,sigma,tg){};
	normal(const normal &a):ctsdistribution(a){};
	~normal(){};
	double next() {return mu()+sigma()*g.normal();}
};
//
//   lognormal distribution
//
class lognormal: public ctsdistribution {
public :
  ctsdistribution* clone(){return new lognormal(*this);}
	ostream &print(ostream &o) const {
		o << "lognormal("<<mu()<<","<<sigma()<<")";return o;
	}
	void print(void) const {
		cout << "lognormal("<<mu()<<","<<sigma()<<")";
	}
	double mu(void) const {return par[0];}
	double sigma(void) const {return par[1];}
	double log_pdf(double x) {
			double lx = (log(x) -mu()) / sigma();
		return LMY_SQRT_2PI - 0.5 * lx * lx - log(sigma()) - log(x);
	}
	lognormal(double mu=0.0, double sigma=1.,rng &tg=globalrandom ):  ctsdistribution(mu,sigma,tg){};
  lognormal(const lognormal &a):ctsdistribution(a){};

  ~lognormal(){};
 private:
	double next(){return exp(mu()+sigma()*g.normal());};
};
//
// beta distribution
//
class beta: public ctsdistribution {
public :
ctsdistribution* clone(){return new beta(*this);}
	ostream &print(ostream &o) const {
		o << "beta"<<"("<<a()<<","<<b()<<")";
		return o;
	}
	void print(void) const {
		cout << "beta"<<"("<<a()<<","<<b()<<")";
	}
	double a(void) const {return par[0];}
	double b(void) const {return par[1];}
	double log_pdf(double x) {
	return (a()-1.)*log(x)+(b()-1)*log(1.-x)
		-lgamma(a())-lgamma(b())+lgamma(a()+b());
}
	beta(double a=1.0, double b=1.,rng &tg=globalrandom)
		:ctsdistribution(a,b,tg){};
	beta(const beta &a):ctsdistribution(a){};
	~beta(){};
  private:
	double next(void) {
	  double x,y;
	  x=g.rgamma(a(),1.0);
	y=g.rgamma(b(),1.0);
	return x/(x+y);
}
};
//
//  uniform distribution
//
class uniform: public ctsdistribution {
public :
ctsdistribution* clone(){return new uniform(*this);}
	ostream &print(ostream &o) const {
		o << "uniform("<<a()<<","<<b()<<")";return o;
	}
	void print(void) const {
		cout << "uniform("<<a()<<","<<b()<<")";
	}
      	double a(void) const {return par[0];}
	double b(void) const {return par[1];}
	double log_pdf(double x) {
	  return 1./(b()-a());
	}
	uniform(double a=0.0, double b=1.,rng &tg=globalrandom):
  ctsdistribution(a,b,tg){
	  range = b-a;
	}
	uniform(const uniform &a):ctsdistribution(a){range=a.range;};
	uniform &operator=(uniform &a) {
	  if (&a==this) return a;
	  ctsdistribution::operator =(a);
	  range=a.range;
			return *this;
		}
	~uniform(){};
private:
	double range;
	double next(void) {
		return a()+g()*range;
	}
};

class undefctsdist: public ctsdistribution {
public :
ctsdistribution* clone(){return new undefctsdist(*this);}
	bool undefined(){return true;};
    ostream &print(ostream &o) const {o<<"undefined";return o;}
	void print(void) const {cout << "undefined";}
	//	double a(void) const {return par[0];}
	//	double b(void) const {return par[1];}
	double sample(void) {return 0.0;}
	double log_pdf(double x) {
		error e("cannot get log_pdf of an undefined distribution");
		return 0.0;
	}
	undefctsdist(double a=0.0, double b=0.,rng &tg=globalrandom):
	ctsdistribution(a,b,tg){};
	undefctsdist(const undefctsdist &a):ctsdistribution(a){};
	~undefctsdist(){};
private:
	double next() {
		error e("cannot sample from undefined distribution");
		return 1.0;
	}
};
ctsdistribution *scan(istream &in);
ctsdistribution *parscan(istream &in, const char *namestring,char default_val[]);
ostream &operator<<(ostream &o, const ctsdistribution &x);
#endif








